SystemOrganization addCategory: #'OmniST-Languages-LRSTM-Errors'!
SystemOrganization addCategory: #'OmniST-Languages-LRSTM-Model'!
SystemOrganization addCategory: #'OmniST-Languages-POCs'!
SystemOrganization addCategory: #'OmniST-Languages-AmbientTalk'!
SystemOrganization addCategory: #'OmniST-Languages-AmbientTalk-Reflection'!
SystemOrganization addCategory: #'OmniST-Languages-STM-Reflection'!
SystemOrganization addCategory: #'OmniST-Languages-STM'!
SystemOrganization addCategory: #'OmniST-Languages-ActiveObject'!
SystemOrganization addCategory: #'OmniST-Languages-ActiveObject-Reflection'!
SystemOrganization addCategory: #'OmniST-Languages-Clojure'!
SystemOrganization addCategory: #'OmniST-Languages-Clojure-Reflection'!
SystemOrganization addCategory: #'OmniST-Languages-STM-Misc'!

OstDomain subclass: #ATActor
	instanceVariableNames: 'process mailbox terminate name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-AmbientTalk-Reflection'!

!ATActor class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/3/2012 11:46'!
example
	| hello |
	hello := ATHelloWorld new.
	hello sayHello.
	hello saySomething: 'baz'.
	
	hello := ATActor newWith: [ ATHelloWorld new ].
	hello sayHello.
	hello saySomething: 'bar'.
! !

!ATActor class methodsFor: 'constructor' stamp: 'StefanMarr 1/3/2012 12:50'!
newWithInitializer: aBlock
	"Normally, we would return a reference to the domain,
	 but this is not really the way AmbientTalk works,
	 so, here we also return the value returned by the aBlock."
	| domain valueMsg promise |
	domain := self new.
	promise := domain evaluateAsync: aBlock.
	^ promise read.! !

!ATActor methodsFor: 'actor semantics' stamp: 'StefanMarr 1/2/2012 16:50'!
evaluateAsync: aBlock
	"This is an implementation level method.
	 The block is by convention not referring to
	 unsafe shared state."
	| msg |
	msg := ATMessage new.
	msg receiver: aBlock.
	msg setSelector: #value arguments: #().

	mailbox nextPut: msg.
	
	^ msg promise.! !

!ATActor methodsFor: 'initialization' stamp: 'StefanMarr 1/4/2012 13:58'!
initialize
	super initialize.
	
	mailbox := SharedQueue new.
	terminate := false.
	
	process := self spawnHere: [
		[terminate] whileFalse: [
			self processIncomingMessages.
		]].
	process name: 'Actor'.! !

!ATActor methodsFor: 'debugging' stamp: 'StefanMarr 1/4/2012 13:43'!
name
	^ name! !

!ATActor methodsFor: 'debugging' stamp: 'StefanMarr 1/4/2012 13:58'!
name: aString
	name := aString.
	
	process ifNotNil: [ process name: 'Actor [', aString, ']' ].! !

!ATActor methodsFor: 'omni primitive handlers' stamp: 'StefanMarr 1/3/2012 16:09'!
primAt: idx on: anObj
	"The message was already send, so we are already in the correct actor."
	<omniDoNotTransform>
	^ anObj at: idx! !

!ATActor methodsFor: 'omni primitive handlers' stamp: 'StefanMarr 1/3/2012 16:09'!
primAt: idx on: anObj put: aVal
	"The message was already send, so we are already in the correct actor."
	<omniDoNotTransform>
	^ anObj at: idx put: aVal! !

!ATActor methodsFor: 'omni primitive handlers' stamp: 'StefanMarr 1/3/2012 16:10'!
primBasicAt: idx on: anObj
	"The message was already send, so we are already in the correct actor."
	<omniDoNotTransform>
	^ anObj basicAt: idx! !

!ATActor methodsFor: 'omni primitive handlers' stamp: 'StefanMarr 1/3/2012 16:09'!
primBasicAt: idx on: anObj put: aVal
	"The message was already send, so we are already in the correct actor."
	<omniDoNotTransform>
	^ anObj basicAt: idx put: aVal! !

!ATActor methodsFor: 'omni primitive handlers' stamp: 'StefanMarr 1/3/2012 16:10'!
primInstVarAt: idx on: anObj
	"The message was already send, so we are already in the correct actor."
	<omniDoNotTransform>
	^ anObj instVarAt: idx! !

!ATActor methodsFor: 'omni primitive handlers' stamp: 'StefanMarr 1/3/2012 16:10'!
primInstVarAt: idx on: anObj put: aVal
	"The message was already send, so we are already in the correct actor."
	<omniDoNotTransform>
	^ anObj instVarAt: idx put: aVal! !

!ATActor methodsFor: 'printing' stamp: 'StefanMarr 1/4/2012 13:54'!
printOn: aStream
	super printOn: aStream.
	
	name ifNotNil: [
		aStream nextPutAll: '[', name, ']'.].! !

!ATActor methodsFor: 'actor semantics' stamp: 'StefanMarr 1/4/2012 14:09'!
processIncomingMessages
	| msg result |
	msg := mailbox next.
	
	result := msg send.
	
	msg result: result.! !

!ATActor methodsFor: 'omni handlers' stamp: 'StefanMarr 1/3/2012 16:11'!
readField: idx of: obj
	<omniDoNotTransform>
	"We rely on the semantics of Smalltalk here.
	 All methods are executed in the AT domain."
	^ obj instVarAt: idx! !

!ATActor methodsFor: 'omni handlers' stamp: 'StefanMarr 1/4/2012 12:58'!
requestExecutionOf: aSelector lookup: cls on: anObj
	<omniDoNotTransform>
	"SEE #requestExecutionOf:on:with: for the template implementation"
	| msg |
	self == OstDomain currentDomain ifTrue: [
		^ anObj performSuper: aSelector inLookupClass: cls.
	].
	
	^ self sendAsAsyncMessage: aSelector to: anObj with: #() inLookupClass: cls.! !

!ATActor methodsFor: 'omni handlers' stamp: 'StefanMarr 1/6/2012 01:03'!
requestExecutionOf: aSelector lookup: cls on: anObj with: p1
	<omniDoNotTransform>
	"SEE #requestExecutionOf:on:with: for the template implementation"
	| msg |
	self == OstDomain currentDomain ifTrue: [
		^ anObj performSuper: aSelector with: {p1} inLookupClass: cls.
	].
	
	^ self sendAsAsyncMessage: aSelector to: anObj with: {p1} inLookupClass: cls.! !

!ATActor methodsFor: 'omni handlers' stamp: 'StefanMarr 1/3/2012 16:11'!
requestExecutionOf: aSelector on: anObj
	<omniDoNotTransform>
	"SEE #requestExecutionOf:on:with: for the template implementation"
	| msg |
	self == OstDomain currentDomain ifTrue: [
		^ anObj perform: aSelector.
	].
	
	^ self sendAsAsyncMessage: aSelector to: anObj with: #().! !

!ATActor methodsFor: 'omni handlers' stamp: 'StefanMarr 1/3/2012 16:17'!
requestExecutionOf: aSelector on: anObj with: p1
	<omniDoNotTransform>
	"All messages sent to objects inside another AmbientTalk domain
	 have to be executed asynchronously, by default,
	 all messages sent inside an actor/domain are executed
	 synchronously. (sync-execution results in direct returning of the result, no future/promise)
	
	 In AT, they do normally return nil, but here we do return always
	 the promise/future. I think, this is more consistent, especially
	 with the way how I handle creation of actors/domains.
	 
	 But, to be able to have simple stuff without promises returned,
	 we use the <noReturnValue> pragma."
	| msg currentDomain |
	currentDomain := OstDomain currentDomain.
	
	self == currentDomain ifTrue: [
		^ anObj perform: aSelector with: p1.
	].
	
	msg := ATMessage new.
	msg setSelector: aSelector arguments: { p1 }.
	msg receiver: anObj.
	mailbox nextPut: msg.
	
	"WARNING: we do not support #noReturnValue at the moment, STEFAN 2012-01-02."
	^ msg result.
	
	"Transcript log: 'message sent. Returns promise: ', 
		(aCompiledMethod pragmas contains: [:p | p keyword = #noReturnValue])
			not asString."
	
	"(aCompiledMethod pragmas contains: [:p | p keyword = #noReturnValue])
		ifTrue:  [^ nil ]
		ifFalse: [^ msg promise]."
! !

!ATActor methodsFor: 'omni handlers' stamp: 'StefanMarr 1/3/2012 16:12'!
requestExecutionOf: aSelector on: anObj with: p1 with: p2
	<omniDoNotTransform>
	"SEE #requestExecutionOf:on:with: for the template implementation"
	| msg |
	self == OstDomain currentDomain ifTrue: [
		^ anObj perform: aSelector with: p1 with: p2.
	].
	
	^ self sendAsAsyncMessage: aSelector to: anObj with: { p1. p2 }.! !

!ATActor methodsFor: 'omni handlers' stamp: 'StefanMarr 1/3/2012 16:12'!
requestExecutionOf: aSelector on: anObj with: p1 with: p2 with: p3
	<omniDoNotTransform>
	"SEE #requestExecutionOf:on:with: for the template implementation"
	| msg |
	self == OstDomain currentDomain ifTrue: [
		^ anObj perform: aSelector with: p1 with: p2 with: p3.
	].
	
	^ self sendAsAsyncMessage: aSelector to: anObj with: { p1. p2. p3 }.! !

!ATActor methodsFor: 'omni handlers' stamp: 'StefanMarr 1/3/2012 16:12'!
requestExecutionOf: aSelector on: anObj with: p1 with: p2 with: p3 with: p4
	<omniDoNotTransform>
	"SEE #requestExecutionOf:on:with: for the template implementation"
	| msg |
	self == OstDomain currentDomain ifTrue: [
		^ anObj perform: aSelector with: p1 with: p2 with: p3 with: p4.
	].
	
	^ self sendAsAsyncMessage: aSelector to: anObj with: { p1. p2. p3. p4 }.! !

!ATActor methodsFor: 'omni handlers' stamp: 'StefanMarr 1/3/2012 16:12'!
requestExecutionOf: aSelector on: anObj with: p1 with: p2 with: p3 with: p4 with: p5
	<omniDoNotTransform>
	"SEE #requestExecutionOf:on:with: for the template implementation"
	| msg |
	self == OstDomain currentDomain ifTrue: [
		^ anObj perform: aSelector with: p1 with: p2 with: p3 with: p4 with: p5.
	].
	
	^ self sendAsAsyncMessage: aSelector to: anObj with: { p1. p2. p3. p4. p5 }.! !

!ATActor methodsFor: 'omni handlers' stamp: 'StefanMarr 1/3/2012 16:12'!
requestExecutionOf: aSelector on: anObj with: p1 with: p2 with: p3 with: p4 with: p5 with: p6
	<omniDoNotTransform>
	"SEE #requestExecutionOf:on:with: for the template implementation"
	| msg |
	self == OstDomain currentDomain ifTrue: [
		^ anObj perform: aSelector with: p1 with: p2 with: p3 with: p4 with: p5 with: p6.
	].
	
	^ self sendAsAsyncMessage: aSelector to: anObj with: { p1. p2. p3. p4. p5. p6 }.! !

!ATActor methodsFor: 'omni handlers' stamp: 'StefanMarr 1/3/2012 16:12'!
requestExecutionOf: aSelector on: anObj with: p1 with: p2 with: p3 with: p4 with: p5 with: p6 with: p7
	<omniDoNotTransform>
	"SEE #requestExecutionOf:on:with: for the template implementation"
	| msg |
	self == OstDomain currentDomain ifTrue: [
		^ anObj perform: aSelector with: p1 with: p2 with: p3 with: p4 with: p5 with: p6 with: p7.
	].
	
	^ self sendAsAsyncMessage: aSelector to: anObj with: { p1. p2. p3. p4. p5. p6. p7 }.! !

!ATActor methodsFor: 'omni handlers' stamp: 'StefanMarr 1/3/2012 16:12'!
requestExecutionOf: aSelector on: anObj with: p1 with: p2 with: p3 with: p4 with: p5 with: p6 with: p7 with: p8
	<omniDoNotTransform>
	| msg |
	self == OstDomain currentDomain ifTrue: [
		^ anObj perform: aSelector with: p1 with: p2 with: p3 with: p4 with: p5 with: p6 with: p7 with: p8.
	].
	
	^ self sendAsAsyncMessage: aSelector to: anObj with: { p1. p2. p3. p4. p5. p6. p7. p8 }.! !

!ATActor methodsFor: 'omni handlers' stamp: 'StefanMarr 1/3/2012 16:12'!
requestExecutionOf: aSelector on: anObj with: p1 with: p2 with: p3 with: p4 with: p5 with: p6 with: p7 with: p8 with: p9
	<omniDoNotTransform>
	| msg |
	self == OstDomain currentDomain ifTrue: [
		^ anObj perform: aSelector with: p1 with: p2 with: p3 with: p4 with: p5 with: p6 with: p7 with: p8 with: p9.
	].
	
	^ self sendAsAsyncMessage: aSelector to: anObj with: { p1. p2. p3. p4. p5. p6. p7. p8. p9 }.! !

!ATActor methodsFor: 'omni handlers' stamp: 'StefanMarr 1/3/2012 16:12'!
requestExecutionOf: aSelector on: anObj with: p1 with: p2 with: p3 with: p4 with: p5 with: p6 with: p7 with: p8 with: p9 with: p10
	<omniDoNotTransform>
	"SEE #requestExecutionOf:on:with: for the template implementation"
	| msg |
	self == OstDomain currentDomain ifTrue: [
		^ anObj perform: aSelector with: p1 with: p2 with: p3 with: p4 with: p5 with: p6 with: p7 with: p8 with: p9 with: p10.
	].
	
	^ self sendAsAsyncMessage: aSelector to: anObj with: { p1. p2. p3. p4. p5. p6. p7. p8. p9. p10 }.! !

!ATActor methodsFor: 'omni handlers' stamp: 'StefanMarr 1/3/2012 16:12'!
requestExecutionOf: aSelector on: anObj with: p1 with: p2 with: p3 with: p4 with: p5 with: p6 with: p7 with: p8 with: p9 with: p10 with: p11
	<omniDoNotTransform>
	| msg |
	self == OstDomain currentDomain ifTrue: [
		^ anObj perform: aSelector with: p1 with: p2 with: p3 with: p4 with: p5 with: p6 with: p7 with: p8 with: p9 with: p10 with: p11.
	].
	
	^ self sendAsAsyncMessage: aSelector to: anObj with: { p1. p2. p3. p4. p5. p6. p7. p8. p9. p10. p11 }.! !

!ATActor methodsFor: 'actor semantics' stamp: 'StefanMarr 1/3/2012 14:44'!
sendAsAsyncMessage: aSelector to: anObj with: anArgArray
	| msg |
	msg := ATMessage new.
	msg setSelector: aSelector arguments: anArgArray.
	msg receiver: anObj.
	mailbox nextPut: msg.
	
	^ msg result.! !

!ATActor methodsFor: 'omni handlers' stamp: 'StefanMarr 1/3/2012 16:12'!
write: val toField: idx of: obj
	<omniDoNotTransform>
	"We rely on Smalltalk's semantics here, this can not be called from another
	 domain since all messages are execute in the AT domain"
	obj instVarAt: idx put: val.
	^ val! !

ATActor subclass: #ATTracingActor
	instanceVariableNames: 'cnt msgTrace'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-AmbientTalk-Reflection'!

!ATTracingActor methodsFor: 'initialization' stamp: 'StefanMarr 1/4/2012 14:56'!
initialize
	cnt := 0.
	msgTrace := OrderedCollection new.
	
	super initialize.! !

!ATTracingActor methodsFor: 'tracing' stamp: 'StefanMarr 1/3/2012 23:54'!
numberOfProcessedMessages
	^ cnt! !

!ATTracingActor methodsFor: 'actor semantics' stamp: 'StefanMarr 1/4/2012 14:55'!
processIncomingMessages
	| msg result |
	msg := mailbox next.
	
	result := msg send.
	
	msgTrace addLast: msg selector.
	cnt := cnt + 1.
	
	msg result: result.
	! !

OstDomain subclass: #ATIsolateDomain
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-AmbientTalk-Reflection'!

OstDomain subclass: #LRSTMDomain
	instanceVariableNames: ''
	classVariableNames: 'Domain'
	poolDictionaries: ''
	category: 'OmniST-Languages-LRSTM-Model'!

!LRSTMDomain class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/11/2011 22:19'!
domain
	"We only need one domain for the STM system."
	Domain ifNil: [ Domain := self new. ].
	^ Domain! !

!LRSTMDomain methodsFor: 'omni primitive handlers' stamp: 'StefanMarr 1/3/2012 15:36'!
primAt: idx on: anObj
	<omniDoNotTransform>
	^ anObj workingCopy at: idx! !

!LRSTMDomain methodsFor: 'omni primitive handlers' stamp: 'StefanMarr 1/3/2012 15:37'!
primAt: idx on: anObj put: anObject
	<omniDoNotTransform>
	^ anObj workingCopy at: idx put: anObject! !

!LRSTMDomain methodsFor: 'omni primitive handlers' stamp: 'StefanMarr 1/3/2012 15:37'!
primBasicAt: idx on: anObj
	<omniDoNotTransform>
	^ anObj workingCopy basicAt: idx! !

!LRSTMDomain methodsFor: 'omni primitive handlers' stamp: 'StefanMarr 1/3/2012 15:37'!
primBasicAt: idx on: anObj put: anObject
	<omniDoNotTransform>
	^ anObj workingCopy basicAt: idx put: anObject! !

!LRSTMDomain methodsFor: 'omni primitive handlers' stamp: 'StefanMarr 1/3/2012 15:56'!
primInstVarAt: idx on: obj
	<omniDoNotTransform>
	^ obj workingCopy instVarAt: idx! !

!LRSTMDomain methodsFor: 'omni primitive handlers' stamp: 'StefanMarr 1/3/2012 16:01'!
primInstVarAt: idx on: obj put: aVal
	<omniDoNotTransform>
	^ obj workingCopy instVarAt: idx put: aVal! !

!LRSTMDomain methodsFor: 'omni handlers' stamp: 'StefanMarr 1/3/2012 15:56'!
readField: idx of: obj
	<omniDoNotTransform>
	^ self primInstVarAt: idx on: obj! !

!LRSTMDomain methodsFor: 'omni handlers' stamp: 'StefanMarr 12/11/2011 22:26'!
readGlobal: association
	<omniDoNotTransform>
	^ association atomicValue! !

!LRSTMDomain methodsFor: 'omni handlers' stamp: 'StefanMarr 1/3/2012 16:02'!
write: val toField: idx of: obj
	<omniDoNotTransform>
	^ self primInstVarAt: idx on: obj put: val.! !

!LRSTMDomain methodsFor: 'omni handlers' stamp: 'StefanMarr 12/11/2011 22:40'!
writeGlobal: association value: val
	<omniDoNotTransform>
	^ association atomicValue: val! !

OstDomain subclass: #OPActiveObjectDomain
	instanceVariableNames: 'obj process mailbox terminate'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-ActiveObject-Reflection'!

!OPActiveObjectDomain class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/5/2011 20:26'!
newFor: anObject
	| domain |
	domain := self new.
	domain object: anObject.
	^ domain.! !

!OPActiveObjectDomain methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 21:04'!
domainForNewObjects
	"Active objects are created explicitly,
	 classes tell whether their objects are supposed to be active.
	 This happens during #initialize.
	 
	 However, #domainForNewObjects is determined during
	 #basicNew, and for ActiveObjects we just assume as the standard
	 case that all objects are passive.
	 For the moment that means, they are not part of any domain."
	^ nil! !

!OPActiveObjectDomain methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/4/2011 17:01'!
initialize
	super initialize.
	
	mailbox := SharedQueue new.
	terminate := false.
	
	process := self spawnHere: [
		[terminate] whileFalse: [
			self processIncomingMessages.
		]].! !

!OPActiveObjectDomain methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/5/2011 20:27'!
object: anObject
	obj := anObject! !

!OPActiveObjectDomain methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/4/2012 14:10'!
processIncomingMessages
	| msg result |
	msg := mailbox next.
	
	result := msg method valueWithReceiver: msg receiver arguments: msg arguments.
	
	msg result: result.! !

!OPActiveObjectDomain methodsFor: 'omni handlers' stamp: 'StefanMarr 11/4/2011 17:16'!
readField: idx of: ignoredObjUseKnownOwnerInstead
	"No strong encapsulation here.
	 Will just read the data and return it."
	^ obj instVarAt: idx! !

!OPActiveObjectDomain methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 20:47'!
requestExecutionOf: aCompiledMethod on: anObj with: arguments
	"For the active object, we want to execute the message asynchronously.
	 But since we can't tell whether something is a void method,
	 we will always need to block on the promise and wait until
	 it is resolved.

	 Another detail: the handling of self sends is also done asynchronously.
    This can be changed, depending on what is desired, but going with this
    approach makes bugs and conceptual inconsitency easier to spot.
    For more complex applications, that is problably not a good idea.
   
    For instance, I ran into trouble with object allocation, and requesting
    the omniDomain from objects by this. Asking for an omniDomain is a reflective
    action and needs to be done via the domain object, which also acts as a mirror."
	| msg |
	Transcript log: 'requestExecutionOf:'.
	msg := OActiveObjectMessage method: aCompiledMethod on: anObj with: arguments.
	mailbox nextPut: msg.
	
	Transcript log: 'message sent. Immediate return: ', (aCompiledMethod pragmas contains: [:p | p keyword = #noReturnValue]) asString.
	
	(aCompiledMethod pragmas contains: [:p | p keyword = #noReturnValue])
		ifTrue:  [^ anObj ]
		ifFalse: [^ msg resolve]! !

!OPActiveObjectDomain methodsFor: 'omni handlers' stamp: 'StefanMarr 11/4/2011 17:16'!
write: val toField: idx of: ignoredObjUseKnownOwnerInstead
	"No strong encapsulation here, will just set the field and return the value."
	obj instVarAt: idx put: val.
	^ val! !

OstDomain subclass: #OPAtomDomain
	instanceVariableNames: 'obj mtx'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-Clojure-Reflection'!

!OPAtomDomain class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 01:42'!
newFor: anObject
	| domain |
	domain := self new.
	domain object: anObject.
	^ domain.! !

!OPAtomDomain methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 21:46'!
domainForNewObjects
	"If I find a way to have these objects in an immutable domain, they should go there.
	 I could keep track of a epoch counter and based on that decide that they are modifiable during
	 the same epoch in which they were created!!
	TODO look into this"
	^ nil! !

!OPAtomDomain methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 01:32'!
initialize
	mtx := Mutex new.! !

!OPAtomDomain methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 01:42'!
object: o
	obj := o! !

!OPAtomDomain methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 01:30'!
readField: idx of: ignoredObjUseKnownOwnerInstead
	"No strong encapsulation here.
	 Will just read the data and return it."
	^ obj instVarAt: idx! !

!OPAtomDomain methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 01:46'!
requestExecutionOf: aCompiledMethod on: anObj with: arguments
	"For the active object, we want to execute the message asynchronously.
	 But since we can't tell whether something is a void method,
	 we will always need to block on the promise and wait until
	 it is resolved."
	| requiresMtx sel |
	Transcript log: 'requestExecutionOf:'.
	
	sel := aCompiledMethod selector.
	
	requiresMtx :=  (sel = #compareTo:andSet:) or: [sel = #reset:].
	
	requiresMtx
		ifTrue:  [ mtx critical: [
			^ aCompiledMethod valueWithReceiver: obj arguments: arguments.]]
		ifFalse: [
			^ aCompiledMethod valueWithReceiver: obj arguments: arguments.].! !

!OPAtomDomain methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 01:46'!
write: val toField: idx of: ignoredObjUseKnownOwnerInstead
	"No strong encapsulation here, will just set the field and return the value."
	obj instVarAt: idx put: val.
	^ val! !

OstDomain subclass: #OPSTMDomain
	instanceVariableNames: 'mtx runningTransactions commitMtx'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-STM-Reflection'!

!OPSTMDomain class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 14:28'!
newWith: initializerBlock
	"Normally, we would return a reference to the domain,
	 but this is not really the way AmbientTalk works,
	 so, here we also return the value returned by the initializerBlock."
	| domain valueMsg promise |
	domain := self new.
	valueMsg := initializerBlock class >> #value.
	promise := domain requestExecutionOf: valueMsg on: initializerBlock with: #().
	^ promise read.! !

!OPSTMDomain methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 18:32'!
commitProtected: theCommitBlock
	^ commitMtx critical: theCommitBlock! !

!OPSTMDomain methodsFor: 'stm' stamp: 'StefanMarr 11/7/2011 15:22'!
currentTransaction
	| proc |
	proc := Processor thisProcess.
	mtx critical: [
		^ runningTransactions at: proc ifAbsent: [nil].
	].! !

!OPSTMDomain methodsFor: 'stm' stamp: 'StefanMarr 11/7/2011 15:21'!
currentTransaction: transaction
	| proc |
	proc := Processor thisProcess.
	mtx critical: [
		| parent |
		parent := runningTransactions at: proc ifAbsent: [nil].
		transaction parent: parent.
		runningTransactions at: proc put: transaction.
	].! !

!OPSTMDomain methodsFor: 'initialization' stamp: 'StefanMarr 11/7/2011 18:45'!
initialize
	super initialize.
	
	mtx := Mutex new.
	runningTransactions := IdentityDictionary new.
	
	commitMtx := Mutex new.! !

!OPSTMDomain methodsFor: 'omni handlers' stamp: 'StefanMarr 11/7/2011 18:11'!
readField: idx of: obj
	" 1. The read has to look in the transaction's write buffer.
	  2. If that failed, it needs to look for the current value in the read buffer.
	  3. If that failed, it needs to look into the actual object."
	| transaction |
	transaction := self currentTransaction.
	transaction
		ifNil: [
			^ obj instVarAt: idx]
		ifNotNil: [
			^ transaction readField: idx of: obj.].
	
! !

!OPSTMDomain methodsFor: 'omni handlers' stamp: 'StefanMarr 11/7/2011 15:24'!
requestExecutionOf: aCompiledMethod on: anObj with: arguments
	"My STM does not have any semantics with regard to messages at the moment.
	 So, lets just do inplace execution"
	^ aCompiledMethod valueWithReceiver: anObj arguments: arguments.! !

!OPSTMDomain methodsFor: 'omni handlers' stamp: 'StefanMarr 11/7/2011 18:13'!
write: val toField: idx of: obj
	"Delegate to the current transaction"
	| transaction |
	transaction := self currentTransaction.
	transaction
		ifNil: 	[^obj instVarAt: idx put: val.]
		ifNotNil: [
			Transcript log: 'write', val asString, ' toField: ', idx asString, ' of: ', obj asString.
			^ self currentTransaction write: val toField: idx of: obj.].! !

OstDomain subclass: #OPSTMHello
	instanceVariableNames: 'omniDomain a b'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-STM'!

!OPSTMHello class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 18:09'!
example
	| stmDomain p1 p2 hello |
	stmDomain := OPSTMDomain new.
	hello := OPSTMHello new.
	
	"Since this main thread is not running inside the OPSTMDomain, which would be easy, too,
	 but is not happening here, we need to manually move the object into the domain."
	stmDomain adoptObject: hello.
	
	hello a: 'main'; b: 'main'.
	
	p1 := stmDomain spawnHere: [
		OPTransaction atomic: [
			hello a: 'p1'.
			Processor yield.
			hello b: 'p1'.
	]].

	p2 := stmDomain spawnHere: [
		OPTransaction atomic: [
			| b |
			b := hello b.
			Processor yield.
			hello b: 'p2'.
	]].! !

!OPSTMHello methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 14:40'!
a
	^ a asString! !

!OPSTMHello methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 14:40'!
a: obj
	a := obj asString! !

!OPSTMHello methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 18:18'!
b
	Transcript log: 'Read b: ', b asString.
	^ b asString! !

!OPSTMHello methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 18:14'!
b: obj
	Transcript log: 'Write b: ', obj asString.
	b := obj asString! !

OstDomain subclass: #OPTransaction
	instanceVariableNames: 'theOmniDomain parent wasAborted writeBuffer readBuffer'
	classVariableNames: 'NoValue'
	poolDictionaries: ''
	category: 'OmniST-Languages-STM-Reflection'!

!OPTransaction class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 17:41'!
NoValue
	^ NoValue! !

!OPTransaction class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/11/2011 18:41'!
atomic: aBlock
	| domain transaction result |
	domain := OstDomain currentDomain.
	
	"We just do it without a transaction if the current domain
	 is not transactional.
	 TODO: make sure that is a sane semantics, especially
	       when interacting with other domains.
	"
	(domain isKindOf: OPSTMDomain) ifFalse: [
		^ aBlock value].
	
	transaction := self new.
	transaction domain: domain.
	domain currentTransaction: transaction.
	
	"Any error that is not catched before this point is "
	result := aBlock ifError: [:err :rcvr |
		"TODO: make sure that this signals an error"
		transaction abortWith: err on: rcvr.].
	
	transaction aborted
		ifTrue: [Error new messageText: 'we should never reach this point for an aborted transaction.'; signal].
	
	"Will also issue an error in case of conflicts.
	 No automatic retry here either."
	transaction commit.
	
	^ result! !

!OPTransaction class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 17:23'!
initialize
	super initialize.
	NoValue := Object new.! !

!OPTransaction methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 18:46'!
abort
	wasAborted := true.
	Error new messageText: 'Transaction aborted.' ; signal! !

!OPTransaction methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 15:40'!
aborted
	^ wasAborted! !

!OPTransaction methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 22:53'!
commit
	theOmniDomain commitProtected: [
		readBuffer checkConsistencyWithObjectsOtherwise: [self abort].
		writeBuffer performWrites.
	].

	Transcript log: 'Completed transaction successfully'.! !

!OPTransaction methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 14:55'!
domain: domain
	" I use a different name for the instance variable here,
	  since I am not sure whether I want to omni-fy these objects,
	  probably not, since it feels like reflective objects.
	  "
	theOmniDomain := domain! !

!OPTransaction methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 17:53'!
initialize
	super initialize.
	
	wasAborted := false.
	
	readBuffer := OPSTMBuffer new.
	writeBuffer := OPSTMBuffer new.! !

!OPTransaction methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 15:39'!
parent: aTransaction
	parent := aTransaction! !

!OPTransaction methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 17:26'!
readField: idx of: obj
	" 1. The read has to look in the transaction's write buffer.
	  2. If that failed, it needs to look for the current value in the read buffer.
	  3. If that failed, it needs to look into the actual object."
	| val |
	val := writeBuffer for: obj instVarAt: idx.
	val == NoValue ifFalse: [^ val].
	
	val := readBuffer for: obj instVarAt: idx.
	val == NoValue ifFalse: [^ val].
	
	val := obj instVarAt: idx.
	
	readBuffer for: obj instVarAt: idx put: val.
	
	^ val.! !

!OPTransaction methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 18:44'!
write: val toField: idx of: obj
	"REM: I was thinking whether to read a previously unread value to guarantee consistency,
	      but after discussion with Joeri, I decided against it.
	 So:
	 1. put the written value in the write buffer.
	 2. keep fingers crossed ;)"
	writeBuffer for: obj instVarAt: idx put: val.
	^ val! !

OstDomain subclass: #OcjAgentDomain
	instanceVariableNames: 'agent'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-Clojure-Reflection'!

!OcjAgentDomain class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/6/2012 13:09'!
newFor: anAgent
	| domain |
	domain := self new.
	domain agent: anAgent.
	^ domain.! !

!OcjAgentDomain methodsFor: 'accessing' stamp: 'StefanMarr 1/5/2012 14:42'!
agent: anAgent
	agent := anAgent! !

!OcjAgentDomain methodsFor: 'omni handlers' stamp: 'StefanMarr 1/6/2012 14:19'!
requestExecutionOf: aSelector on: anObject
	"An agent supports three synchronous operations with no parameters:
	  #await
	  #read
	  #shutdown"
	<omniDoNotTransform>
	| normSel |
	
	"Rules are only enforced on the agent itself"
	anObject = agent ifFalse: [
		^ anObject perform: aSelector].
	
	normSel := aSelector asNormalSelector.
	
	(normSel = #read or:   [
	 normSel = #await or:  [
	 normSel = #shutdown   ]])
		ifTrue: [
		^ agent perform: normSel	].
	
	"There are two more selectors, we do not want to have someone messing with:"
	(normSel = #processIncomingMessages or: [
	 normSel = #initialize])
		ifTrue: [
			Error signal: 'Access to implementation methods is denied.'].
	
	"For everything else, we just assume someone added an explicit protocal to a Smalltalk subclass.
	 Thus, we follow Agent semantics by doing it asynchronously, but we do not give a return value.
	 Could be done with a promise, but we will encorage here asynchnous/agent-like programming."
	^ agent perform: #send: with: [:oldState | agent perform: aSelector.
														oldState].! !

!OcjAgentDomain methodsFor: 'omni handlers' stamp: 'StefanMarr 1/6/2012 14:44'!
requestExecutionOf: aSelector on: anObject with: p1
	"An agent supports one synchronous operations with one parameters:
	  #send:"
	<omniDoNotTransform>
	| normSel |	
	anObject = agent ifFalse: [
		^ anObject perform: aSelector with: p1.].

	
	normSel := aSelector asNormalSelector.
	
	(normSel = #send:) ifTrue: [
		^ agent send: p1].
	
	"For everything else, we just assume someone added an explicit protocal to a Smalltalk subclass.
	 Thus, we follow Agent semantics by doing it asynchronously, but we do not give a return value.
	 Could be done with a promise, but we will encorage here asynchnous/agent-like programming."
	
	^ agent perform: #send: with: [:oldState | agent perform: aSelector with: p1.
														oldState].! !

!OcjAgentDomain methodsFor: 'omni handlers' stamp: 'StefanMarr 1/6/2012 14:19'!
requestExecutionOf: aSelector on: anObject with: p1 with: p2
	"An agent supports one synchronous operations with two parameters:
	  #send:with:"
	<omniDoNotTransform>
	| normSel |
	anObject = agent ifFalse: [
		^ anObject perform: aSelector with: p1 with: p2.].
	
	normSel := aSelector asNormalSelector.
	
	(normSel = #send:with:) ifTrue: [
		^ agent send: p1 with: p2].
	
	"For everything else, we just assume someone added an explicit protocal to a Smalltalk subclass.
	 Thus, we follow Agent semantics by doing it asynchronously, but we do not give a return value.
	 Could be done with a promise, but we will encorage here asynchnous/agent-like programming."
	
	^ agent perform: #send:with: with: [:oldState | agent perform: aSelector with: p1 with: p2.
															  oldState].! !

OstDomain subclass: #OcjImmutableDomain
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-Clojure-Reflection'!

!OcjImmutableDomain methodsFor: 'omni primitive handlers' stamp: 'StefanMarr 1/6/2012 15:11'!
primAt: idx on: anObj put: aVal
	<omniDoNotTransform>
	self raiseImmutabilityError.! !

!OcjImmutableDomain methodsFor: 'omni primitive handlers' stamp: 'StefanMarr 1/6/2012 15:12'!
primBasicAt: idx on: anObj put: aVal
	<omniDoNotTransform>
	self raiseImmutabilityError.! !

!OcjImmutableDomain methodsFor: 'omni primitive handlers' stamp: 'StefanMarr 1/6/2012 15:12'!
primInstVarAt: idx on: anObj put: aVal
	<omniDoNotTransform>
	self raiseImmutabilityError.! !

!OcjImmutableDomain methodsFor: 'immutability' stamp: 'StefanMarr 1/6/2012 15:14'!
raiseImmutabilityError
	Error signal: 'Attempt to modify the object failed. The owning domain does not allow mutations.'! !

!OcjImmutableDomain methodsFor: 'omni handlers' stamp: 'StefanMarr 1/6/2012 15:12'!
write: val toField: idx of: obj
	<omniDoNotTransform>
	self raiseImmutabilityError.! !

!OcjImmutableDomain methodsFor: 'omni handlers' stamp: 'StefanMarr 1/6/2012 15:12'!
writeGlobal: association value: val
	<omniDoNotTransform>
	self raiseImmutabilityError.! !

!ContextPart methodsFor: '*OmniST-Languages-LRSTM' stamp: 'lr 5/31/2007 15:49'!
workingCopy
	^ self! !

Message subclass: #ATMessage
	instanceVariableNames: 'promise receiverObj'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-AmbientTalk-Reflection'!

!ATMessage class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 22:52'!
method: aCompiledMethod on: anObj with: arguments
	| msg |
	msg := self new.
	msg method: aCompiledMethod.
	msg receiver: anObj.
	msg arguments: arguments.
	^ msg! !

!ATMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/4/2012 14:11'!
initialize
	super initialize.
	promise := OPromise new.! !

!ATMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 22:53'!
promise
	^ promise! !

!ATMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 22:52'!
receiver
	^ receiverObj! !

!ATMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 22:52'!
receiver: anObj
	receiverObj := anObj! !

!ATMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/2/2012 16:52'!
result
	"Result is blocking on the promise.
	 Use #promise to obtain the promise directly."
	^ promise read! !

!ATMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 22:52'!
result: anObj
	promise deliver: anObj! !

!ATMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/2/2012 16:53'!
send
	"answer the result of sending this message to its receiver"

	^ receiverObj perform: selector withArguments: args! !

!WriteStream methodsFor: '*OmniST-Languages-LRSTM' stamp: 'StefanMarr 12/4/2011 22:17'!
atomicNextPut: anObject
	((collection class == ByteString) and: [
		anObject isCharacter and:[anObject isOctetCharacter not]]) ifTrue: [
			collection := (WideString from: collection).
			^self nextPut: anObject.
	].
	position >= writeLimit
		ifTrue: [^ self pastEndPut: anObject]
		ifFalse: 
			[position := position + 1.
			^collection at: position put: anObject]! !

OstObject subclass: #ATHelloWorld
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-AmbientTalk'!

!ATHelloWorld methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 12:41'!
sayHello
	Transcript log: 'Hello World from within an actor. Actor Domain:', omniDomain printString, '(', omniDomain hash asString, ')'.
	self saySomething: 'foooo'.! !

!ATHelloWorld methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 12:40'!
saySomething: aString
	Transcript log: 'Something: ', aString, '. Actor Domain:', omniDomain printString, '(', omniDomain hash asString, ')'.! !

OstObject subclass: #OcjAgent
	instanceVariableNames: 'state mailbox terminate process'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-Clojure'!

!OcjAgent class methodsFor: 'instance creation' stamp: 'StefanMarr 1/6/2012 13:27'!
new: initialState
	"Work around omni enforcement"
	<omniDoNotTransform>
	
	| agent stateIdx |
	agent := self new.
	agent instVarNamed: #state put: initialState.
	
	^ agent.! !

!OcjAgent methodsFor: 'agent' stamp: 'StefanMarr 1/6/2012 13:15'!
await
	"Block until all actions have been processed (weaker semantics than Clojure's)"
	<omniDoNotTransform>
	[mailbox isEmpty] 
		whileFalse: [
			(Delay forMilliseconds: 50) wait.
		]! !

!OcjAgent methodsFor: 'initialization' stamp: 'StefanMarr 1/6/2012 13:23'!
initialize
	<omniDoNotTransform>
	super initialize.
	
	mailbox := SharedQueue new.
	terminate := false.
	
	omniDomain := OcjAgentDomain newFor: self.
	process := omniDomain newProcessFor: [
		[terminate and: [mailbox isEmpty]] whileFalse: [
			self processIncomingMessages.
		]].
	process name: 'OcjAgent'.
	process resume.! !

!OcjAgent methodsFor: 'agent' stamp: 'StefanMarr 1/7/2012 12:38'!
processIncomingMessages
	| updateBlock |
	updateBlock := mailbox peekWaiting. "Wait until we have it."

	state := updateBlock value: state.
	
	mailbox next. "Removing it only now avoids races with await"! !

!OcjAgent methodsFor: 'agent' stamp: 'StefanMarr 1/6/2012 13:16'!
read
	<omniDoNotTransform>
	^ state! !

!OcjAgent methodsFor: 'agent' stamp: 'StefanMarr 1/6/2012 13:17'!
send: anUpdateBlock
	<omniDoNotTransform>
	mailbox nextPut: anUpdateBlock.! !

!OcjAgent methodsFor: 'agent' stamp: 'StefanMarr 1/6/2012 13:17'!
send: anUpdateBlock with: args
	<omniDoNotTransform>
	self send: [:old | anUpdateBlock value: old value: args ]! !

!OcjAgent methodsFor: 'agent' stamp: 'StefanMarr 1/6/2012 13:17'!
shutdown
	"Make sure that the process loop gets around to see that terminate is set to true.
	 So, we just set it ourselves.
	 No race conditions here, purely sequential."
	<omniDoNotTransform>
	mailbox nextPut: [:ignore | terminate := true].! !

OcjAgent subclass: #OcjAgentWithImmutability
	instanceVariableNames: 'immutabilityDomain'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-Clojure'!

!OcjAgentWithImmutability methodsFor: 'initialization' stamp: 'StefanMarr 1/6/2012 15:15'!
initialize
	<omniDoNotTransform>
	super initialize.
	immutabilityDomain := OcjImmutableDomain new.! !

!OcjAgentWithImmutability methodsFor: 'agent' stamp: 'StefanMarr 1/6/2012 15:00'!
processIncomingMessages
	super processIncomingMessages.
	immutabilityDomain adoptObject: state.! !

!Process methodsFor: '*OmniST-Languages-LRSTM' stamp: 'lr 5/30/2007 12:09'!
currentTransaction
	^ currentTransaction! !

!Process methodsFor: '*OmniST-Languages-LRSTM' stamp: 'lr 5/30/2007 12:09'!
currentTransaction: aTransaction
	currentTransaction := aTransaction! !

Error subclass: #LRSTMConflict
	instanceVariableNames: 'transaction'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-LRSTM-Errors'!

!LRSTMConflict methodsFor: 'accessing' stamp: 'lr 5/25/2007 09:04'!
transaction
	^ transaction! !

!LRSTMConflict methodsFor: 'accessing' stamp: 'lr 5/25/2007 09:04'!
transaction: aTransaction
	transaction := aTransaction! !

Error subclass: #LRSTMInvalidTransaction
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-LRSTM-Errors'!

!BlockContext methodsFor: '*OmniST-Languages-LRSTM' stamp: 'lr 2/26/2008 18:16'!
atomic
	^ LRSTMTransaction new do: self! !

!BlockContext methodsFor: '*OmniST-Languages-LRSTM' stamp: 'lr 5/25/2007 09:13'!
atomicIfConflict: aBlock
	^ LRSTMTransaction new do: self ifConflict: aBlock! !

!BlockContext methodsFor: '*OmniST-Languages-LRSTM' stamp: 'lr 5/25/2007 09:14'!
atomicRetry
	^ LRSTMTransaction new retry: self! !

!ReadStream methodsFor: '*OmniST-Languages-LRSTM' stamp: 'lr 6/7/2007 15:16'!
atomicNext
	^ position >= readLimit
		ifFalse: [ collection at: (position := position + 1) ]! !

!BlockClosure methodsFor: '*OmniST-Languages-LRSTM' stamp: 'StefanMarr 11/28/2011 20:58'!
atomic
	^ LRSTMTransaction new do: self! !

!BlockClosure methodsFor: '*OmniST-Languages-LRSTM' stamp: 'lr 5/25/2007 09:14'!
atomicIfConflict: aBlock
	^ LRSTMTransaction new do: self ifConflict: aBlock! !

!BlockClosure methodsFor: '*OmniST-Languages-LRSTM' stamp: 'lr 5/25/2007 09:15'!
atomicRetry
	^ LRSTMTransaction new retry: self! !

!LookupKey methodsFor: '*OmniST-Languages-LRSTM' stamp: 'lr 5/25/2007 11:18'!
atomicValue
	^ self workingCopy value! !

!LookupKey methodsFor: '*OmniST-Languages-LRSTM' stamp: 'lr 5/25/2007 11:18'!
atomicValue: anObject
	^ self workingCopy value: anObject! !

Object subclass: #LRSTMChange
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-LRSTM-Model'!

!LRSTMChange methodsFor: 'actions' stamp: 'lr 6/8/2007 16:59'!
apply
	self subclassResponsibility! !

!LRSTMChange methodsFor: 'testing' stamp: 'lr 6/8/2007 17:00'!
hasChanged
	^ true! !

!LRSTMChange methodsFor: 'testing' stamp: 'lr 6/8/2007 17:00'!
hasConflict
	^ false! !

LRSTMChange subclass: #LRSTMCustomChange
	instanceVariableNames: 'applyBlock conflictTestBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-LRSTM-Model'!

!LRSTMCustomChange class methodsFor: 'instance-creation' stamp: 'StefanMarr 12/11/2011 16:33'!
onApply: anApplyBlock
	<omniDoNotTransform>
	^ self onApply: anApplyBlock onConflictTest: [ false ]! !

!LRSTMCustomChange class methodsFor: 'instance-creation' stamp: 'StefanMarr 12/11/2011 16:33'!
onApply: anApplyBlock onConflictTest: aConflictTestBlock
	<omniDoNotTransform>
	^ self basicNew initializeApply: anApplyBlock conflictTest: aConflictTestBlock! !

!LRSTMCustomChange methodsFor: 'actions' stamp: 'lr 2/26/2008 16:26'!
apply
	applyBlock value! !

!LRSTMCustomChange methodsFor: 'testing' stamp: 'lr 6/8/2007 17:28'!
hasConflict
	^ conflictTestBlock value! !

!LRSTMCustomChange methodsFor: 'initialization' stamp: 'lr 6/8/2007 17:04'!
initializeApply: anApplyBlock conflictTest: aConflictTestBlock	
	applyBlock := anApplyBlock.
	conflictTestBlock := aConflictTestBlock! !

LRSTMChange subclass: #LRSTMObjectChange
	instanceVariableNames: 'original previous working'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-LRSTM-Model'!

!LRSTMObjectChange class methodsFor: 'instance-creation' stamp: 'lr 6/8/2007 16:59'!
on: anObject
	^ self basicNew initializeOn: anObject! !

!LRSTMObjectChange methodsFor: 'actions' stamp: 'lr 6/8/2007 16:59'!
apply
	original restoreSnapshot: working! !

!LRSTMObjectChange methodsFor: 'testing' stamp: 'lr 6/8/2007 17:00'!
hasChanged
	^ (working isIdenticalToSnapshot: previous) not! !

!LRSTMObjectChange methodsFor: 'testing' stamp: 'lr 6/8/2007 17:00'!
hasConflict
	^ (original isIdenticalToSnapshot: previous) not! !

!LRSTMObjectChange methodsFor: 'initialization' stamp: 'lr 6/8/2007 16:59'!
initializeOn: anObject
	original := anObject.
	working := original snapshotCopy.
	previous := working snapshotCopy! !

!LRSTMObjectChange methodsFor: 'accessing' stamp: 'lr 6/8/2007 17:00'!
original
	^ original! !

!LRSTMObjectChange methodsFor: 'accessing' stamp: 'lr 6/8/2007 17:00'!
previous
	^ previous! !

!LRSTMObjectChange methodsFor: 'printing' stamp: 'lr 6/8/2007 17:01'!
printOn: aStream
	super printOn: aStream.
	aStream nextPutAll: ' ['; print: self original; nextPut: $]! !

!LRSTMObjectChange methodsFor: 'accessing' stamp: 'lr 6/8/2007 17:00'!
working
	^ working! !

Object subclass: #LRSTMTransaction
	instanceVariableNames: 'parent context changes objectChanges'
	classVariableNames: ''
	poolDictionaries: 'ContextPart'
	category: 'OmniST-Languages-LRSTM-Model'!

!LRSTMTransaction class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/11/2011 22:01'!
initialize
	super initialize.
	
	(Process hasInstVarNamed: #currentTransaction) ifFalse: [
		Process addInstVarNamed: #currentTransaction.
	].! !

!LRSTMTransaction class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/11/2011 22:02'!
unload
	(Process hasInstVarNamed: #currentTransaction) ifTrue: [
		Process removeInstVarNamed: #currentTransaction.
	]! !

!LRSTMTransaction methodsFor: 'private' stamp: 'StefanMarr 12/11/2011 16:28'!
abort
	"Abort a transaction."
	
	<omniDoNotTransform>
	parent := context := changes := objectChanges := nil! !

!LRSTMTransaction methodsFor: 'public' stamp: 'StefanMarr 12/11/2011 16:29'!
abort: aValue
	"Abort a transaction."
	
	<omniDoNotTransform>
	Processor activeProcess currentTransaction: parent.
	thisContext swapSender: context.
	self abort.
	^ aValue! !

!LRSTMTransaction methodsFor: 'accessing' stamp: 'StefanMarr 12/11/2011 16:29'!
addChange: aChange
	<omniDoNotTransform>
	^ changes add: aChange! !

!LRSTMTransaction methodsFor: 'private' stamp: 'StefanMarr 12/11/2011 16:31'!
begin
	"Start a transaction. If this is a nested transaction copy the parent changes into our own context."
	
	<omniDoNotTransform>
	changes := OrderedCollection new.
	objectChanges := IdentityDictionary new.
	self isNested ifTrue: [
		parent objectChanges
			keysAndValuesDo: [ :key :value | objectChanges at: key put: value copy ] ]! !

!LRSTMTransaction methodsFor: 'accessing' stamp: 'StefanMarr 12/11/2011 16:32'!
changeFor: anObject
	<omniDoNotTransform>
	^ objectChanges at: anObject ifAbsentPut: [ self addChange: (LRSTMObjectChange on: anObject) ]! !

!LRSTMTransaction methodsFor: 'accessing' stamp: 'StefanMarr 12/11/2011 16:32'!
changes
	"Answer an ordered collection of unique changes."

	<omniDoNotTransform>
	^ changes! !

!LRSTMTransaction methodsFor: 'public' stamp: 'StefanMarr 12/11/2011 16:32'!
checkpoint
	"Checkpoint a transaction."

	<omniDoNotTransform>
	self commit; begin! !

!LRSTMTransaction methodsFor: 'private' stamp: 'StefanMarr 12/11/2011 16:32'!
commit
	"Commit a transaction."

	<omniDoNotTransform>
	self commitIfConflict: [ self signalConflict ]! !

!LRSTMTransaction methodsFor: 'private' stamp: 'StefanMarr 12/11/2011 16:32'!
commitIfConflict: aBlock
	"Commit a transaction atomically."
	
	<omniDoNotTransform>
	[	changes do: [ :each |
			each hasConflict
				ifTrue: [ ^ aBlock value ] ].
		self isNested 
			ifTrue: [
				objectChanges keysAndValuesDo: [ :key :value |
					(parent objectChanges includesKey: key)
						ifFalse: [ parent addChange: value ].
					parent objectChanges at: key put: value ] ]
			ifFalse: [ 	
				changes do: [ :each |
					each hasChanged
						ifTrue: [ each apply ] ] ] ]
		valueUnpreemptively! !

!LRSTMTransaction methodsFor: 'public' stamp: 'StefanMarr 12/11/2011 16:32'!
do: aBlock
	"Start a transaction with aBlock."

	<omniDoNotTransform>		
	^ self do: aBlock ifConflict: [ self signalConflict ]! !

!LRSTMTransaction methodsFor: 'public' stamp: 'StefanMarr 12/11/2011 22:18'!
do: aBlock ifConflict: aConflictBlock
	"Evaluate aBLock within a new transaction, unless we are already in an existing transactional context. Evaluate aConflictBlock if the transaction conflicts with concurrent edits."

	<omniDoNotTransform>
	| result |
	parent := Processor activeProcess currentTransaction.
	Processor activeProcess currentTransaction: self.
	context := thisContext sender.
	self begin.
	
	[[ result := aBlock 
		ensure: [ Processor activeProcess currentTransaction: parent ] ]
		ifCurtailed: [ self abort ].
	] omni: LRSTMDomain domain.
 
	self commitIfConflict: aConflictBlock.
	^ result! !

!LRSTMTransaction methodsFor: 'testing' stamp: 'StefanMarr 12/11/2011 16:33'!
isNested
	<omniDoNotTransform>
	^ parent notNil! !

!LRSTMTransaction methodsFor: 'accessing' stamp: 'StefanMarr 12/11/2011 16:33'!
objectChanges
	<omniDoNotTransform>
	^ objectChanges! !

!LRSTMTransaction methodsFor: 'accessing' stamp: 'StefanMarr 12/11/2011 16:33'!
parent
	<omniDoNotTransform>
	^ parent! !

!LRSTMTransaction methodsFor: 'public' stamp: 'StefanMarr 12/11/2011 16:35'!
retry: aBlock
	"Unconditionally retry to evaluate aBlock until there are no conflict."

	<omniDoNotTransform>
	^ self do: aBlock ifConflict: [ self retry: aBlock ]! !

!LRSTMTransaction methodsFor: 'private' stamp: 'StefanMarr 12/11/2011 16:35'!
signalConflict
	<omniDoNotTransform>
	^ LRSTMConflict new transaction: self; signal! !

Object subclass: #OActiveObjectMessage
	instanceVariableNames: 'promise methodToBeExecuted receiverObj arguments'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-ActiveObject-Reflection'!

!OActiveObjectMessage class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/5/2011 17:17'!
method: aCompiledMethod on: anObj with: arguments
	| msg |
	msg := self new.
	msg method: aCompiledMethod.
	msg receiver: anObj.
	msg arguments: arguments.
	^ msg! !

!OActiveObjectMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/5/2011 23:17'!
arguments
	^ arguments! !

!OActiveObjectMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/5/2011 17:26'!
arguments: argArray
	arguments := argArray! !

!OActiveObjectMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/4/2011 17:23'!
initialize
	super initialize.
	promise := OPromise new.! !

!OActiveObjectMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/5/2011 23:17'!
method
	^ methodToBeExecuted! !

!OActiveObjectMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/5/2011 17:19'!
method: aCompiledMethod
	methodToBeExecuted := aCompiledMethod! !

!OActiveObjectMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/5/2011 23:17'!
receiver
	^ receiverObj! !

!OActiveObjectMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/5/2011 17:19'!
receiver: anObj
	receiverObj := anObj! !

!OActiveObjectMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/4/2011 17:24'!
resolve
	^ promise read.! !

!OActiveObjectMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/5/2011 23:19'!
result: anObj
	promise deliver: anObj! !

Object subclass: #OPActiveObject
	instanceVariableNames: 'omniDomain'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-ActiveObject'!

!OPActiveObject class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/5/2011 20:22'!
example
	"When instantianting an active object,
	 it does everything to be able to transform itself to a real active object.
	 Thus, all messages are handled asynchronously, and for that an event loop is
	 created, waiting on a certain message queue.
	 
	 Beyond that, we do not give any other interesting guarantees.
	 
	 Since we don't have anyother way of determining whether a method actually
	 returns a useful value, we use an explicit pragma:
	   <noReturnValue>
	 This indicates that the method does not return anything useful,
	 and the client does not block, but gets a self instead."! !

!OPActiveObject methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/5/2011 20:25'!
initialize
	omniDomain := OPActiveObjectDomain newFor: self.! !

OPActiveObject subclass: #OpAOPingPong
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-ActiveObject'!

!OpAOPingPong class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 00:47'!
example
	[ | ping |
		ping := OpAOPingPong new.
		ping ping. ] forkNamed: 'ping-pong-test'.! !

!OpAOPingPong methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 20:04'!
ping
	<noReturnValue>
	Transcript log: 'ping'.
	Transcript log: 'PingPong domain: ', omniDomain printString, '(', omniDomain hash asString, ')'.
	
	OpAOPingObj new printDomain.
	
	self pong.
	! !

!OpAOPingPong methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/5/2011 23:26'!
pong
	<noReturnValue>
	Transcript addEntry: 'pong'.! !

Object subclass: #OPAgentMessage
	instanceVariableNames: 'methodToBeExecuted arguments'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-Clojure-Reflection'!

!OPAgentMessage class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 02:33'!
method: aCompiledMethod with: arguments
	| msg |
	msg := self new.
	msg method: aCompiledMethod.
	msg arguments: arguments.
	^ msg! !

!OPAgentMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 02:33'!
arguments
	^ arguments! !

!OPAgentMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 02:32'!
arguments: argArray
	arguments := argArray! !

!OPAgentMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 02:32'!
method
	^ methodToBeExecuted! !

!OPAgentMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 02:32'!
method: aCompiledMethod
	methodToBeExecuted := aCompiledMethod! !

Object subclass: #OPAtom
	instanceVariableNames: 'omniDomain state'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-Clojure'!

!OPAtom class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 02:03'!
example
	| atom |
	atom := OPAtom new.

	Transcript log: 'New atom: ', atom read asString.

	atom reset: 1.

	[1 to: 10 do: [:i |
		atom swap: [:old | 
			Transcript log: '[atom-1] old: ', old asString, ' in iteration: ', i asString.
			Processor yield.
			old + 1]
	]] forkNamed: 'atom-1'.

	[1 to: 10 do: [:i |
		atom swap: [:old | 
			Transcript log: '[atom-2] old: ', old asString, ' in iteration: ', i asString.
			Processor yield.
			old + 1]
	]] forkNamed: 'atom-2'.
! !

!OPAtom methodsFor: 'atom' stamp: 'StefanMarr 11/6/2011 01:43'!
compareTo: oldValue andSet: newValue
	"Clojure: (compare-and-set!! atom oldval newval)
	 Returns true on success, false otherwise."
	state = oldValue
		ifTrue:  [ state := newValue.
					 ^ true  ]
		ifFalse: [ ^ false ].! !

!OPAtom methodsFor: 'initialization' stamp: 'StefanMarr 11/6/2011 01:47'!
initialize
	omniDomain := OPAtomDomain newFor: self.! !

!OPAtom methodsFor: 'atom' stamp: 'StefanMarr 11/6/2011 01:43'!
read
	"Clojure deref/@ reads the most recent value"
	^ state! !

!OPAtom methodsFor: 'atom' stamp: 'StefanMarr 11/6/2011 01:45'!
reset: value
	state := value! !

!OPAtom methodsFor: 'atom' stamp: 'StefanMarr 11/6/2011 01:57'!
swap: updateBlock
	"Clojure: (swap!! atom f)
	 The block needs to take an argument,
	 and return a result value, like this:
	 [:oldVal | oldVal + 5]	"
	| currentState |
	[ | result |
	  currentState := state.
	  result := self compareTo: currentState
			andSet: (updateBlock value: currentState).
	  result ifFalse: [Transcript log: 'swap failed'].
	  result
		] whileFalse.
	
	! !

Object subclass: #OPSTMBuffer
	instanceVariableNames: 'map'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-STM-Misc'!

!OPSTMBuffer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 22:53'!
checkConsistencyWithObjectsOtherwise: abortBlock
	"Execute abortBlock if buffer is not consistent with the referenced objects"
	| NoValue |
	NoValue := OPTransaction NoValue.
	map keysAndValuesDo: [:obj :shadow |
		shadow withIndexDo: [:val :idx |
			val ~= NoValue ifTrue: [
				"Ok, that is a value we read,
				 so it needs to be consistent with what we have currently."
				(obj instVarAt: idx) == val
					"Ooops, not consistent, so we abort here."
					ifFalse: [^abortBlock value.]
				]
			]
		].! !

!OPSTMBuffer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 17:44'!
for: obj instVarAt: idx
	| shadow |
	shadow := map at: obj
					 ifAbsent: [^ OPTransaction NoValue].
	^ shadow at: idx.! !

!OPSTMBuffer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 18:23'!
for: obj instVarAt: idx put: val
	| shadow |
	shadow := map at: obj
					 ifAbsentPut: [self newShadowFor: obj].
	^ shadow at: idx put: val.! !

!OPSTMBuffer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 17:39'!
initialize
	super initialize.
	
	map := IdentityDictionary new.! !

!OPSTMBuffer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 17:51'!
newShadowFor: obj
	| n shadow |
	n := obj class numberOfInstanceVariables.
	shadow := Array new: n.
	shadow atAllPut: OPTransaction NoValue.
	^ shadow! !

!OPSTMBuffer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/7/2011 18:41'!
performWrites
	| NoValue |
	NoValue := OPTransaction NoValue.
	map keysAndValuesDo: [:obj :shadow |
		shadow withIndexDo: [:val :idx |
			val ~= NoValue ifTrue: [
				"Ok, that is a value we wrote,
				 now lets write it for real"
				obj instVarAt: idx put: val.
			]
		]
	]! !

!Object methodsFor: '*OmniST-Languages-LRSTM-atomic' stamp: 'StefanMarr 12/11/2011 16:31'!
atomicShallowCopy
	<omniDoNotTransform>
	^ self workingCopy shallowCopy! !

!Object methodsFor: '*OmniST-Languages-LRSTM' stamp: 'lr 6/7/2007 15:05'!
isIdenticalToSnapshot: anObject
	1 to: self class instSize do: [ :index | 
		(self instVarAt: index) == (anObject instVarAt: index)
			ifFalse: [ ^ false ] ].
	self class isVariable ifTrue: [
		1 to: self size do: [ :index | 
			(self at: index) == (anObject at: index)
				ifFalse: [ ^ false ] ] ].
	^ true! !

!Object methodsFor: '*OmniST-Languages-LRSTM' stamp: 'lr 6/7/2007 13:47'!
restoreSnapshot: anObject
	<primitive: 168>
	self primitiveFailed! !

!Object methodsFor: '*OmniST-Languages-LRSTM' stamp: 'lr 6/7/2007 13:48'!
snapshotCopy
	<primitive: 148>
	self primitiveFailed! !

!Object methodsFor: '*OmniST-Languages-LRSTM' stamp: 'lr 5/30/2007 12:11'!
workingCopy
	"Answer a working copy to be used within the context of the current transaction."
	
	| transaction |
	transaction := Processor activeProcess currentTransaction
		ifNil: [ self error: 'No active transaction' ].
	^ (transaction changeFor: self) working! !

Object subclass: #OcjAgentExample
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-Clojure'!

!OcjAgentExample class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/7/2012 12:40'!
example
	| agent |
	agent := OcjAgent new: 1.

	[ 1 to: 10 do: [:i|
		agent send: [:old |
			Transcript log: '[agent-1] old: ', old asString, ' iteration: ', i asString.
			old + 1].
		(Delay forSeconds: 1) wait.
	]] forkNamed: 'agent-1'.

	[ 1 to: 10 do: [:i|
		agent send: [:old |
			Transcript log: '[agent-2] old: ', old asString, ' iteration: ', i asString.
			old + 1].
		(Delay forMilliseconds: 800) wait.
	]] forkNamed: 'agent-2'.

	[ 1 to: 20 do: [:i|
		agent send: [:old |
			Transcript log: '[agent-3] old: ', old asString, ' iteration: ', i asString.
			old + 1].
		(Delay forMilliseconds: 512) wait.
	]] forkNamed: 'agent-3'.

	Transcript log: 'Started all agent tester processes.'.! !

Object subclass: #OpAOPingObj
	instanceVariableNames: 'omniDomain'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-ActiveObject'!

!OpAOPingObj methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 19:59'!
printDomain
	Transcript log: omniDomain printString, '(', omniDomain hash asString, ')'.! !

Object subclass: #OstExperiments
	instanceVariableNames: 'omniDomain a b c d'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Languages-POCs'!

!OstExperiments commentStamp: '<historical>' prior: 0!
This class had only a use while developing the compiler transformations.!

!OstExperiments methodsFor: 'nil' stamp: 'StefanMarr 12/7/2011 23:19'!
a: val
	a := val! !

!OstExperiments methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/7/2011 23:20'!
a: anA b: aB andSome: aC
	a := anA + aB + aC! !

!OstExperiments methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/7/2011 22:22'!
aAndB
	^ a + b! !

!OstExperiments methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/7/2011 23:20'!
abc
 | sd |
 a ifNil: [^b].
 a == b ifTrue: [^ a].
 ^ c! !

!OstExperiments methodsFor: 'nil' stamp: 'StefanMarr 12/8/2011 00:33'!
do: p1 a: p2 b: p3 c: p4 d: p5 e: p6
	^ p1 asString,
	  p2 asString,
	  p3 asString,
	  p4 asString,
	  p5 asString,
	  p6 asString
	  ! !

!OstExperiments methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/9/2011 13:44'!
manualTestChainedSend
	| result |
	
	result := [ 
		| x_omni_1_ |
		x_omni_1_ := 1 omniDomain requestExecutionOf: #bitOr: on: 1 with: 2.
		x_omni_1_ omniDomain requestExecutionOf: #bitOr: on: x_omni_1_ with: 4
	] fooBarEntryMessage.

	self omniDomain
		requestExecutionOf: #xx_omni_x_assert:equals:
		on: self
		with: 7
		with: result! !

!OstExperiments methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/8/2011 00:31'!
perform: sel a: p1 b: p2 c: p3 d: p4 e: p5 f: p6
	"dd"
	<primitive: 83>
	^ -1! !
LRSTMTransaction initialize!
OPTransaction initialize!
